#include <powerbox.h>

AsyncWebServer webServer(HTTP_PORT);
AsyncWebSocket webSocket("/ws");

PowerOutlet powerOutlets[NUMBERPOWEROUTLETS];
Ticker autostartTickers[NUMBERPOWEROUTLETS];
bool touchDetected[NUMBERPOWEROUTLETS];

WifiConfig wifiConfig;

void setup()
{
  // Debug console
  Serial.begin(115200);
  while (!Serial)
    ;

  delay(200);

  // Start the filesystem
  LittleFS.begin();

  // Load de WiFi configuration
  loadWiFiConfiguration();
  setupWifi();

  // Initialize the GPIO
  setupGPIO();

  // Initialize the autostart up tickers
  initializeTickers();

  setupWebserver();

  setupTouchInterfaces();
}

bool setupWebserver()
{

  // serve all files from /
  webServer.serveStatic("/", LittleFS, "/").setDefaultFile("index.html");

  // REST API endpoints
  // get /power_outlets/{id}
  webServer.on("^\\/power_outlets\\/([0-9]+)$", HTTP_GET, processGetPowerOutlets_Id);
  // put /power_outlets/{id}
  webServer.on(
      "^\\/power_outlets\\/([0-9]+)$", HTTP_PUT, [](AsyncWebServerRequest *request) {},
      NULL, processPutPowerOutlets_Id);

  // get /power_outlets
  webServer.on("/power_outlets", HTTP_GET, processGetPowerOutlets);

  // get /wifi_config
  webServer.on("/wifi_config", HTTP_GET, processGetWifiConfig);

  // /test
  AsyncCallbackJsonWebHandler *wifiConfigHandler =
      new AsyncCallbackJsonWebHandler("/wifi_config", processHttpWifiConfig);

  webServer.addHandler(wifiConfigHandler);

  // notFound page
  webServer.onNotFound(processNotFound);

  // enable websocket
  webSocket.onEvent(onEvent);
  webServer.addHandler(&webSocket);

  // start the webserver
  webServer.begin();

  return true;
}

void loop()
{
  webSocket.cleanupClients();
}

void processGetPowerOutlets(AsyncWebServerRequest *request)
{
  Serial << F("webserver: processGetPowerOutlets") << endl;

  AsyncResponseStream *response = request->beginResponseStream("application/json");
  StaticJsonDocument<768> responseDocument;

  for (int i = 0; i < NUMBERPOWEROUTLETS; i++)
  {
    responseDocument.add(powerOutlets[i].toJson());
  }

  serializeJsonPretty(responseDocument, Serial);

  serializeJson(responseDocument, *response);

  request->send(response);
}

void processGetPowerOutlets_Id(AsyncWebServerRequest *request)
{
  int powerOutletId = request->pathArg(0).toInt();

  Serial << F("webserver: processGetPowerOutlets of id: ") << powerOutletId << endl;

  AsyncResponseStream *response = request->beginResponseStream("application/json");

  serializeJson(powerOutlets[powerOutletId].toJson(), *response);

  request->send(response);
}

void processGetWifiConfig(AsyncWebServerRequest *request)
{

  Serial << F("Webserver: processGetWifiConfig") << endl;

  AsyncResponseStream *response = request->beginResponseStream("application/json");

  serializeJson(wifiConfig.toJson(), *response);

  request->send(response);
}

void processHttpWifiConfig(AsyncWebServerRequest *request, JsonVariant &json)
{
  Serial << F("Webserver: processHttpWifiConfig - ") << request->methodToString() << endl;

  switch (request->method())
  {
  // HTTP_PUT
  case HTTP_PUT:
  {

    IPAddress tempIPAddress;
    char tempIPAddressChar[15];

    AsyncResponseStream *response = request->beginResponseStream("application/json");

    serializeJsonPretty(json, Serial);

    wifiConfig.setWifiNetWorkSSid(json["wifi_network_ssid"]);
    wifiConfig.setWifiNetWorkPassword(json["wifi_network_password"]);
    wifiConfig.setSoftAPSsid(json["soft_ap_ssid"]);
    wifiConfig.setSoftAPPassword(json["soft_ap_password"]);
    strlcpy(tempIPAddressChar, json["soft_ap_ip_address"],
            sizeof(tempIPAddressChar));
    tempIPAddress.fromString(tempIPAddressChar);
    wifiConfig.setSoftAPIPAddress(tempIPAddress);
    strlcpy(tempIPAddressChar, json["soft_ap_subnet"],
            sizeof(tempIPAddressChar));
    tempIPAddress.fromString(tempIPAddressChar);
    wifiConfig.setSoftAPSubnet(tempIPAddress);
    strlcpy(tempIPAddressChar, json["soft_ap_gateway"],
            sizeof(tempIPAddressChar));
    tempIPAddress.fromString(tempIPAddressChar);
    wifiConfig.setSoftAPGateway(tempIPAddress);

    wifiConfig.persist(wifiConfigurationFile);

    serializeJson(wifiConfig.toJson(), *response);

    request->send(response);

    break;
  }

  // HTTP_GET
  case HTTP_GET:
  {
    AsyncResponseStream *response = request->beginResponseStream("application/json");

    serializeJson(wifiConfig.toJson(), *response);

    request->send(response);
    break;
  }

  default:
  {
    request->send(404, "text/plain", "Request not supported");
    break;
  }
  }
}

void processNotFound(AsyncWebServerRequest *request)
{
  request->send(404, "text/plain", "Not found");
}

void processPutPowerOutlets_Id(AsyncWebServerRequest *request, uint8_t *data, size_t len, size_t index, size_t total)
{
  int powerOutletId = request->pathArg(0).toInt();

  Serial << F("webserver: processPutPowerOutlets of id: ") << powerOutletId << endl;

  if (!index)
  {
    Serial.printf("BodyStart: %u B\n", total);
  }
  for (size_t i = 0; i < len; i++)
  {
    Serial.write(data[i]);
  }
  if (index + len == total)
  {
    Serial.printf("BodyEnd: %u B\n", total);
  }

  AsyncResponseStream *response = request->beginResponseStream("application/json");
  StaticJsonDocument<96> document;
  DeserializationError error = deserializeJson(document, (char *)data);

  if (document["id"] == powerOutletId)
  {
    powerOutlets[powerOutletId].setName(document["name"]);

    savePowerOutletsConfiguration();

    serializeJson(powerOutlets[powerOutletId].toJson(), *response);
  }
  else
  {
  }
  request->send(response);
}

void setupGPIO()
{
  Serial << F("Setting up GPIO") << endl;

  // Load the initial relay configuration
  loadPowerOutletsConfiguration();

  for (int i = 0; i < NUMBERPOWEROUTLETS; i++)
  {
    int pin = powerOutlets[i].getPin();
    pinMode(pin, OUTPUT);
    digitalWrite(pin, LOW);
  }
}

bool loadPowerOutletsConfiguration()
{
  // Read the pins configuration
  File file = LittleFS.open("/pinConfig.json");
  // Allocate a temporary JsonDocument
  StaticJsonDocument<384> pinsDoc;
  // Deserialize the JSON document
  DeserializationError error = deserializeJson(pinsDoc, file);
  if (error)
    Serial << F("Failed to read pinConfig.json") << endl;
  file.close();

  // Read the power outlet data
  file = LittleFS.open(powerOutletsConfigFile);
  // Allocate a temporary JsonDocument
  StaticJsonDocument<768> powerOutletsDoc;
  // Deserialize the JSON document
  error = deserializeJson(powerOutletsDoc, file);
  if (error)
    Serial << F("Failed to read powerOutletsConfig.json, using default configuration") << endl;
  file.close();

  // Copy values from the JsonDocument
  for (int i = 0; i < NUMBERPOWEROUTLETS; i++)
  {
    powerOutlets[i].setId(powerOutletsDoc[i]["id"]);
    powerOutlets[i].setName(powerOutletsDoc[i]["name"]);
    powerOutlets[i].setAutostart(powerOutletsDoc[i]["autostart"]);
    powerOutlets[i].setDelay(powerOutletsDoc[i]["delay"]);
    powerOutlets[i].setPin(powerOutletsDoc[i]["pin"]);
    // retrieve the pin information
    for (int j = 0; j < NUMBERPOWEROUTLETS; j++)
      if (pinsDoc[j]["pin"] == powerOutletsDoc[i]["pin"])
      {
        powerOutlets[i].setVoltage(pinsDoc[j]["voltage"]);
        powerOutlets[i].setConnector(pinsDoc[j]["connector"]);
        powerOutlets[i].setTouchPin(pinsDoc[j]["touch_pin"]);

        // break out of for-loop
        j = NUMBERPOWEROUTLETS;
      }
  }

  Serial << F("Initial outlet states:") << endl;
  for (int i = 0; i < NUMBERPOWEROUTLETS; i++)
  {
    Serial << F("\t") << powerOutlets[i].getName() << F(":\t") << powerOutlets[i].getState();
    Serial << F("\t") << powerOutlets[i].getPin() << F("\t") << powerOutlets[i].getVoltage();
    Serial << F("\t") << powerOutlets[i].getConnector() << endl;
  }

  return true;
}

bool savePowerOutletsConfiguration()
{
  File file = LittleFS.open(powerOutletsConfigFile, "w");

  StaticJsonDocument<512> saveDocument;

  for (int i = 0; i < NUMBERPOWEROUTLETS; i++)
  {
    Serial << endl
           << endl;
    saveDocument.add(powerOutlets[i].toJson());
  }

  serializeJsonPretty(saveDocument, Serial);
  Serial << endl
         << endl;
  serializeJsonPretty(saveDocument, file);

  file.close();

  return true;
}

bool loadWiFiConfiguration()
{
  File file = LittleFS.open(wifiConfigurationFile);

  IPAddress tempIPAddress;
  char tempIPAddressChar[15];

  // Allocate a temporary JsonDocument
  StaticJsonDocument<512> doc;

  // Deserialize the JSON document
  DeserializationError error = deserializeJson(doc, file);

  if (error)
    Serial << F("Failed to read file, using default configuration") << endl;

  // Copy values from the JsonDocument to the wifiConfig class
  wifiConfig.setWifiNetWorkSSid(doc["wifi_network_ssid"]);
  wifiConfig.setWifiNetWorkPassword(doc["wifi_network_password"]);
  wifiConfig.setSoftAPSsid(doc["soft_ap_ssid"]);
  wifiConfig.setSoftAPPassword(doc["soft_ap_password"]);
  strlcpy(tempIPAddressChar, doc["soft_ap_ip_address"],
          sizeof(tempIPAddressChar));
  tempIPAddress.fromString(tempIPAddressChar);
  wifiConfig.setSoftAPIPAddress(tempIPAddress);
  strlcpy(tempIPAddressChar, doc["soft_ap_subnet"],
          sizeof(tempIPAddressChar));
  tempIPAddress.fromString(tempIPAddressChar);
  wifiConfig.setSoftAPSubnet(tempIPAddress);
  strlcpy(tempIPAddressChar, doc["soft_ap_gateway"],
          sizeof(tempIPAddressChar));
  tempIPAddress.fromString(tempIPAddressChar);
  wifiConfig.setSoftAPGateway(tempIPAddress);

  file.close();

  serializeJsonPretty(wifiConfig.toJson(), Serial);
  Serial << endl;

  return true;
}

void processTouch()
{
}

void setupTouchInterfaces()
{

  for (int i = 0; i < NUMBERPOWEROUTLETS; i++)
  {
    // Setup the touch interupts
    touchAttachInterrupt(powerOutlets[i].getTouchPin(), processTouch, TOUCH_THRESHOLD);
  }
}

void setupWifi()
{
  // Set up AP
  WiFi.mode(WIFI_MODE_APSTA);

  WiFi.softAPConfig(wifiConfig.getSoftAPIPAddress(), wifiConfig.getSoftAPGateway(), wifiConfig.getSoftAPSubnet());
  WiFi.softAP(wifiConfig.getSoftAPSsid().c_str(), wifiConfig.getSoftAPPassword().c_str());

  // Connect to WiFi
  WiFi.begin(wifiConfig.getWifiNetworkSsid().c_str(), wifiConfig.getWifiNetworkPassword().c_str());

  if (WiFi.waitForConnectResult() != WL_CONNECTED)
  {
    Serial << F("Wifi connection failed!!") << endl;
    return;
  }
  else
  {
    Serial << F("Wifi configuration:") << endl;
    Serial << F("\tIP: ") << WiFi.localIP() << endl;
  }
}

void togglePowerOutlet(int id)
{
  setPowerOutletState(id, !powerOutlets[id].getState());
}

void setPowerOutletState(int id, bool state)
{
  powerOutlets[id].setState(state);

  digitalWrite(powerOutlets[id].getPin(), state ? HIGH : LOW);
  notifyWebSocketClients(id);
}

void turnOffAll()
{
  for (int i = 0; i < NUMBERPOWEROUTLETS; i++)
  {
    setPowerOutletState(i, false);
  }
}

void notifyWebSocketClients(int id)
{
  String message;

  serializeJson(powerOutlets[id].toJson(), message);

  webSocket.textAll(message);
}

void onEvent(AsyncWebSocket *server, AsyncWebSocketClient *client, AwsEventType type,
             void *arg, uint8_t *data, size_t len)
{
  switch (type)
  {
  case WS_EVT_CONNECT:
    Serial << F("WebSocket client #") << client->id() << F(" connected from ") << client->remoteIP().toString() << endl;
    break;
  case WS_EVT_DISCONNECT:
    Serial << F("WebSocket client #") << client->id() << F(" disconnected") << endl;
    break;
  case WS_EVT_DATA:
    handleWebSocketMessage(arg, data, len);
    break;
  case WS_EVT_PONG:
  case WS_EVT_ERROR:
    break;
  }
}

void handleWebSocketMessage(void *arg, uint8_t *data, size_t len)
{
  StaticJsonDocument<32> document;
  AwsFrameInfo *info = (AwsFrameInfo *)arg;

  Serial << F("Handling WS event") << endl;
  if (info->final && info->index == 0 && info->len == len && info->opcode == WS_TEXT)
  {
    // Deserialize the JSON document
    DeserializationError error = deserializeJson(document, (char *)data);

    serializeJsonPretty(document, Serial);
    Serial << endl;

    if (strcmp(document["cmd"], "toggle") == 0)
    {
      Serial << F("WS: toggle power outlet ") << endl;
      togglePowerOutlet(document["id"]);
    }

    if (strcmp(document["cmd"], "turn_off_all") == 0)
    {
      Serial << F("WS: turning all power outlets off") << endl;
      turnOffAll();
    }
  }
}

void initializeTickers()
{
  for (int i = 0; i < NUMBERPOWEROUTLETS; i++)
  {
    if (powerOutlets[i].getAutostart())
    {
      autostartTickers[i].attach(powerOutlets[i].getDelay(), setonPowerOutlet, i);
    }
  }
}

void setonPowerOutlet(int id)
{
  Serial << powerOutlets[id].getName() << F(" autostarted after ") << millis() << endl;
  setPowerOutletState(id, true);
  
  autostartTickers[id].detach();
}